/*
 * Copyright (c) Camden Dixie O'Brien
 * SPDX-License-Identifier: AGPL-3.0-only
 */

#include "parse.h"

#include <assert.h>
#include <stdlib.h>

#define CLASS_START_CAPACITY 4
#define SEQUENCE_START_CAPACITY 8
#define TREE_START_CAPACITY 4

static bool is_special(char c)
{
	switch (c) {
	case '|':
	case '.':
	case '\\':
	case '(':
	case ')':
	case '*':
	case '+':
	case '?':
	case '[':
	case ']':
	case '^':
		return true;
	default:
		return false;
	}
}

static int parse_literal(const char *input, int rem, char *out)
{
	if (rem > 0 && !is_special(input[0])) {
		*out = input[0];
		return 1;
	} else if (rem > 1 && '\\' == input[0]) {
		*out = input[1];
		return 2;
	} else {
		return PARSE_FAIL;
	}
}

static int parse_class(const char *input, int rem, parse_class_t *out)
{
	int result, used = 0;

	if (used >= rem || '[' != input[used])
		return PARSE_FAIL;
	++used;

	if (used < rem && '^' == input[used]) {
		out->negated = true;
		++used;
	} else {
		out->negated = false;
	}

	out->count = 0;
	out->capacity = CLASS_START_CAPACITY;
	out->contents = malloc(out->capacity);
	assert(NULL != out->contents);

	while (used < rem) {
		if (out->count >= out->capacity) {
			out->capacity *= 2;
			out->contents = realloc(out->contents, out->capacity);
			assert(NULL != out->contents);
		}

		result = parse_literal(
		    input + used, rem - used, &out->contents[out->count]);
		if (PARSE_FAIL == result)
			break;
		used += result;
		++out->count;
	}

	if (used >= rem || ']' != input[used])
		return PARSE_FAIL;
	++used;

	return out->count > 0 ? used : -1;
}

static int parse_term(const char *input, int rem, parse_term_t *out)
{
	int result, used = 0;

	if (1 > rem)
		return PARSE_FAIL;

	if ('.' == input[0]) {
		out->type = PARSE_TERM_WILDCARD;
		++used;
	} else if ('(' == input[0]) {
		++used;

		result = parse_expr(input + used, rem - used, &out->subexpr);
		if (PARSE_FAIL == result)
			return PARSE_FAIL;
		out->type = PARSE_TERM_SUBEXPR;
		used += result;

		if (')' != input[used])
			return PARSE_FAIL;
		++used;
	} else if ('[' == input[0]) {
		result = parse_class(input + used, rem - used, &out->class);
		if (PARSE_FAIL == result)
			return PARSE_FAIL;
		out->type = PARSE_TERM_CLASS;
		used += result;
	} else {
		result = parse_literal(input + used, rem - used, &out->literal);
		if (PARSE_FAIL == result)
			return PARSE_FAIL;
		out->type = PARSE_TERM_LITERAL;
		used += result;
	}

	if (used < rem) {
		switch (input[used]) {
		case '*':
			out->quantifier = PARSE_QUANTIFIER_STAR;
			++used;
			break;
		case '+':
			out->quantifier = PARSE_QUANTIFIER_PLUS;
			++used;
			break;
		case '?':
			out->quantifier = PARSE_QUANTIFIER_QMARK;
			++used;
			break;
		default:
			out->quantifier = PARSE_QUANTIFIER_NONE;
		}
	} else {
		out->quantifier = PARSE_QUANTIFIER_NONE;
	}

	return used;
}

static int parse_sequence(const char *input, int rem, parse_sequence_t *out)
{
	int result, used = 0;

	out->count = 0;
	out->capacity = SEQUENCE_START_CAPACITY;
	out->contents = malloc(out->capacity * sizeof(parse_term_t));
	assert(NULL != out->contents);

	while (used < rem) {
		if (out->count >= out->capacity) {
			out->capacity *= 2;
			out->contents = realloc(
			    out->contents, out->capacity * sizeof(parse_term_t));
			assert(NULL != out->contents);
		}

		result = parse_term(
		    input + used, rem - used, &out->contents[out->count]);
		if (PARSE_FAIL == result)
			break;
		++out->count;
		used += result;
	}

	return out->count > 0 ? used : -1;
}

int parse_expr(const char *input, int rem, parse_tree_t *out)
{
	int result, used = 0;

	out->count = 0;
	out->capacity = TREE_START_CAPACITY;
	out->contents = malloc(out->capacity * sizeof(parse_sequence_t));
	assert(NULL != out->contents);

	result = parse_sequence(input + used, rem - used, &out->contents[0]);
	if (PARSE_FAIL == result)
		return PARSE_FAIL;
	++out->count;
	used += result;

	while (used < rem) {
		if ('|' != input[used])
			break;
		++used;

		if (out->count >= out->capacity) {
			out->capacity *= 2;
			out->contents = realloc(
			    out->contents, out->capacity * sizeof(parse_sequence_t));
			assert(NULL != out->contents);
		}

		result = parse_sequence(
		    input + used, rem - used, &out->contents[out->count]);
		if (PARSE_FAIL == result)
			break;
		++out->count;
		used += result;
	}

	return used;
}

static void class_free(const parse_class_t *c)
{
	if (NULL != c->contents)
		free(c->contents);
}

static void sequence_free(const parse_sequence_t *s)
{
	if (NULL != s->contents) {
		for (int i = 0; i < s->count; ++i) {
			switch (s->contents[i].type) {
			case PARSE_TERM_CLASS:
				class_free(&s->contents[i].class);
				break;
			case PARSE_TERM_SUBEXPR:
				parse_tree_free(&s->contents[i].subexpr);
				break;
			default:
				break;
			}
		}
		free(s->contents);
	}
}

void parse_tree_free(const parse_tree_t *t)
{
	if (NULL != t->contents) {
		for (int i = 0; i < t->count; ++i)
			sequence_free(&t->contents[i]);
		free(t->contents);
	}
}
